/*
 * linux/arch/mips/vr41xx/irq.c
 *
 * Code to handle VR41XX IRQs plus some generic interrupt stuff.
 *
 * Copyright (C) 1992 Linus Torvalds
 * Copyright (C) 1994, 1995, 1996, 1997 Ralf Baechle
 * Copyright (C) 1999 Bradley D. LaRonde
 * Copyright (C) 1999, 2000 Michael Klar
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 */

#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>
#include <linux/random.h>
#include <linux/pm.h>

#include <asm/irq.h>
#include <asm/mipsregs.h>
#include <asm/gdb-stub.h>
#include <asm/vr41xx.h>
#include <asm/vr41xx-platdep.h>

extern asmlinkage void vr41xx_handle_irq(void);
extern void breakpoint(void);

// The following two are not used here, but needed elsewhere
irq_cpustat_t irq_stat[NR_CPUS];
unsigned long spurious_count = 0;

static unsigned short irq_mask_probe[(VR41XX_IRQ_MAX + 9)/16];

#define BIT_MASK(bit) ( 1 << (bit) )
#define MAKE_CP0_STATUS_IRQ_MASK(irq) ( BIT_MASK( (irq) + 8))

static inline void mask_irq(unsigned int irq)
{
	if (irq < 8) {
		// it's a cpu interrupt
		unsigned short newstatus = read_32bit_cp0_register(CP0_STATUS);
		newstatus &= ~((unsigned short)1 << (irq + 8)); 
		set_cp0_status(ST0_IM, newstatus);
	} else {
		if (irq < 40) {
			// it's an ICU interrupt
			if (irq < 24) {
				*VR41XX_MSYSINT1REG &= ~((unsigned short)1 << (irq - 8));
			} else {
				*VR41XX_MSYSINT2REG &= ~((unsigned short)1 << (irq - 24));
			}
		} else {
			// it's a GPIO interrupt
#ifdef CONFIG_CPU_VR4181
			*VR41XX_GPINTMSK |= (unsigned short)1 << (irq - 40);
#else
			if (irq < 56) {
				*VR41XX_MGIUINTLREG &= ~((unsigned short)1 << (irq - 40));
			} else {
				*VR41XX_MGIUINTHREG &= ~((unsigned short)1 << (irq - 56));
			}
#endif
		}
	}
}

static inline void unmask_irq(unsigned int irq)
{
	if (irq < 8) {
		// it's a cpu interrupt
		unsigned short newstatus = read_32bit_cp0_register(CP0_STATUS);
		newstatus |= ((unsigned short)1 << (irq + 8)); 
		set_cp0_status(ST0_IM, newstatus);
	} else {
		if (irq < 40) {
			// it's an ICU interrupt
			if (irq < 24) {
				*VR41XX_MSYSINT1REG |= (unsigned short)1 << (irq - 8);
			} else {
				*VR41XX_MSYSINT2REG |= (unsigned short)1 << (irq - 24);
			}
		} else {
			// it's a GPIO interrupt (also ack edge-triggered or hold ints)
#ifdef CONFIG_CPU_VR4181
			if (!((irq < 48 ? *VR41XX_GPINTTYPL : *VR41XX_GPINTTYPH) &
			      ((unsigned short)2 << ((irq & 0x7) * 2))))
				*VR41XX_GPINTSTAT = (unsigned short)1 << (irq - 40);
			*VR41XX_GPINTMSK &= ~((unsigned short)1 << (irq - 40));
#else
			if (irq < 56) {
				if (*VR41XX_GIUINTHTSELL & ((unsigned short)1 << (irq - 40)))
					*VR41XX_GIUINTSTATL = (unsigned short)1 << (irq - 40);
				*VR41XX_MGIUINTLREG |= (unsigned short)1 << (irq - 40);
			} else {
				if (*VR41XX_GIUINTHTSELH & ((unsigned short)1 << (irq - 56)))
					*VR41XX_GIUINTSTATH = (unsigned short)1 << (irq - 56);
				*VR41XX_MGIUINTHREG |= (unsigned short)1 << (irq - 56);
			}
#endif
		}
	}
}

/*
 * Per-IRQ information, by not initializing, this gets filled with NULLs:
 */
typedef struct {
	struct irqaction* irq_action;	// info on low-level handler
	int depth;			// < 0: enabled, >= 0: disabled
} irq_info_t;

static irq_info_t irq_info[VR41XX_IRQ_MAX + 1];

void disable_irq(unsigned int irq_nr)
{
	unsigned long flags;

	if (++irq_info[irq_nr].depth >= 0) {
		save_and_cli(flags);
		mask_irq(irq_nr);
		restore_flags(flags);
	}
}

void enable_irq(unsigned int irq_nr)
{
	unsigned long flags;

#ifdef CONFIG_CPU_VR4122
	if (irq_nr == VR41XX_IRQ_GPIO14)
	    return;
#endif

	if (--irq_info[irq_nr].depth < 0) {
		save_and_cli(flags);
		unmask_irq(irq_nr);
		restore_flags(flags);
	}
}

int get_irq_list(char *buf)
{
	// make a human-readable list of irq actions
	// used by /proc/interrupts

	int i, len = 0;
	struct irqaction *action;

	for (i = 0; i <= VR41XX_IRQ_MAX; i++) {
		action = irq_info[i].irq_action;
		if (!action)
			continue;
		len += sprintf(buf + len, "%2d: %8d %c %s",
			i, kstat.irqs[0][i],
			(action->flags & SA_INTERRUPT) ? '+' : ' ',
			action->name);
		for (action = action->next; action; action = action->next) {
			len += sprintf(buf + len, ",%s %s",
				(action->flags & SA_INTERRUPT) ? " +" : "",
				action->name);
		}
		len += sprintf(buf + len, "\n");
   	 }
	return len;
}

atomic_t __mips_bh_counter;

asmlinkage void do_IRQ(int irq, struct pt_regs *regs)
{
	// This handles all IRQ's that have been installed.
	// It is called from int-handler.S.
	// Actions without SA_INTERRUPT run with interrupts enabled
	// and use the full signal-handling return.
	// Actions with SA_INTERRUPT run with interrupts disabled.

	struct irqaction *action;
	int cpu = smp_processor_id();

	irq_enter(cpu);

	kstat.irqs[cpu][irq]++;

	// don't interrupt on this same irq again until we're finished
	// also, it gets left masked if there is no action (see below)
	mask_irq(irq);

	action = irq_info[irq].irq_action;

	if (action != 0) {
		unsigned long flags = 0;

		if (!(action->flags & SA_INTERRUPT))
			__sti();

		do {
			// handle it
			action->handler(irq, action->dev_id, regs);

			flags |= action->flags;
			action = action->next;
		} while (action);

		if (flags & SA_SAMPLE_RANDOM)
			add_interrupt_randomness(irq);

		__cli();

		unmask_irq(irq);
	}

	irq_exit(cpu);

	// Unmasking and softirq handling is done for us
	// currently by ret_from_irq in entry.S.
}

int add_irq_action(int irq, struct irqaction *new)
{
	// put the specified action (new) at the end of the list of actions for this irq

	int shared = 0;
	struct irqaction *old, **p;
	unsigned long flags;

	p = &irq_info[irq].irq_action;

	if ((old = *p) != NULL) {
		if (!(old->flags & new->flags & SA_SHIRQ))
			return -EBUSY;

		// Shared interrupts must be all same type
		if ((old->flags ^ new->flags) & SA_INTERRUPT)
			return -EBUSY;

		do {
			p = &old->next;
			old = *p;
		} while (old);

		shared = 1;
	}

	if (new->flags & SA_SAMPLE_RANDOM)
		rand_initialize_irq(irq);

	save_and_cli(flags);

	*p = new;

	if (!shared) {
		irq_info[irq].depth = -1;
		unmask_irq(irq);
	}

	restore_flags(flags);

	return 0;
}

int request_irq(unsigned int irq,
	void (*handler)(int, void *, struct pt_regs *),
	unsigned long irqflags,
	const char *devname,
	void *dev_id)
{
	int retval;
	struct irqaction *action;

	// some assertiveness
	if (irq > VR41XX_IRQ_MAX)
		return -EINVAL;

	if (!handler)
		return -EINVAL;

	// allocate a new action struct
	action = (struct irqaction *) kmalloc(sizeof(struct irqaction), GFP_KERNEL);
	if (!action)
		return -ENOMEM;

	// initialize it
	action->handler = handler;
	action->flags = irqflags;
	action->mask = 0;
	action->name = devname;
	action->next = NULL;
	action->dev_id = dev_id;

	// add it to the list of actions for this irq
	retval = add_irq_action(irq, action);
	if (retval)
		kfree(action);

	return retval;
}

void free_irq(unsigned int irq, void *dev_id)
{
	struct irqaction *action, **p;
	unsigned long flags;

	if (irq > VR41XX_IRQ_MAX) {
		printk("Error - trying to free invalid IRQ %d\n", irq);
		return;
	}

	for (p = &irq_info[irq].irq_action; (action = *p) != NULL; p = &action->next) {
		if (action->dev_id != dev_id)
			continue;

		// Found it - now free it
		save_and_cli(flags);
		*p = action->next;
		if (!irq_info[irq].irq_action) {
			irq_info[irq].depth = 0;
			mask_irq(irq);
		}
		restore_flags(flags);
		kfree(action);
		return;
	}
}

unsigned long probe_irq_on(void)
{
	int i;
	unsigned long delay;

	// Note the lower 8 bits of irq_mask_probe[0] are not used. Also,
	// we don't probe for IRQ 0, since no way to report (0 = no IRQ found)
	irq_mask_probe[0] = read_32bit_cp0_register(CP0_STATUS);
	irq_mask_probe[1] = *VR41XX_MSYSINT1REG;
	irq_mask_probe[2] = *VR41XX_MSYSINT2REG;
#ifdef CONFIG_CPU_VR4181
	irq_mask_probe[3] = ~*VR41XX_GPINTMSK;
#else
	irq_mask_probe[3] = *VR41XX_MGIUINTLREG;
	irq_mask_probe[4] = *VR41XX_MGIUINTHREG;
#endif

	for (i = VR41XX_IRQ_MAX; i > 0; i--)
	    if (!irq_info[i].irq_action && i != VR41XX_IRQ_TIMER) {
		enable_irq(i);
	    }

	// Wait for spurious interrupts to mask themselves out again...
	for (delay = jiffies + HZ/10; time_before(jiffies, delay); )
		barrier();	// about 100ms of delay

	irq_mask_probe[0] |= ~read_32bit_cp0_register(CP0_STATUS);
	irq_mask_probe[1] |= ~*VR41XX_MSYSINT1REG;
	irq_mask_probe[2] |= ~*VR41XX_MSYSINT2REG;
#ifdef CONFIG_CPU_VR4181
	irq_mask_probe[3] |= *VR41XX_GPINTMSK;
#else
	irq_mask_probe[3] |= ~*VR41XX_MGIUINTLREG;
	irq_mask_probe[4] |= ~*VR41XX_MGIUINTHREG;
#endif

	return 0x12345678;
}

int probe_irq_off(unsigned long unused)
{
	int i, irq_found, nr_irqs;
	unsigned short tmp;
	unsigned long flags;

	if (unused != 0x12345678)
		printk("Bad IRQ probe detected\n");

	// The following mess unmasks the interrupts we enabled to autoprobe
	// and finishes bit-processing irq_mask_probe at the same time
	save_and_cli(flags);
	tmp = read_32bit_cp0_register(CP0_STATUS);
	set_cp0_status(ST0_IM, tmp & irq_mask_probe[0]);
	irq_mask_probe[0] |= tmp;
	tmp = *VR41XX_MSYSINT1REG;
	*VR41XX_MSYSINT1REG = tmp & irq_mask_probe[1];
	irq_mask_probe[1] |= tmp;
	tmp = *VR41XX_MSYSINT2REG;
	*VR41XX_MSYSINT2REG = tmp & irq_mask_probe[2];
	irq_mask_probe[2] |= tmp;
#ifdef CONFIG_CPU_VR4181
	tmp = ~*VR41XX_GPINTMSK;
	*VR41XX_GPINTMSK = ~(tmp & irq_mask_probe[3]);
	irq_mask_probe[3] |= tmp;
#else
	tmp = *VR41XX_MGIUINTLREG;
	*VR41XX_MGIUINTLREG = tmp & irq_mask_probe[3];
	irq_mask_probe[3] |= tmp;
	tmp = *VR41XX_MGIUINTHREG;
	*VR41XX_MGIUINTHREG = tmp & irq_mask_probe[4];
	irq_mask_probe[4] |= tmp;
#endif
	restore_flags(flags);

	nr_irqs = 0;
	irq_found = 0;

	for (i = VR41XX_IRQ_MAX; i > 0; i--)
		if (!(irq_mask_probe[(i + 8)/16] & ((unsigned short)1 << ((i + 8) & 15)))) {
			irq_found = i;
			nr_irqs++;
		}

	printk("probe_irq_off: nr_irqs %d irq_found %d\n", nr_irqs, irq_found);
	if (nr_irqs > 1)
		irq_found = -irq_found;
	return irq_found;
}

#ifdef CONFIG_PM
//
// Unlike the real pm_request callbacks, this one doesn't get registered
// with PM, and only gets called from do_hibernate and do_wakeup, because
// it has to happen in a certain order.  It also assumes ints disabled.
//
void do_pm_irq_request(pm_request_t rqst)
{
	static unsigned short irq_mask[(VR41XX_IRQ_MAX + 9)/16];
	unsigned int status;

	switch (rqst) {
	case PM_RESUME:
		status = read_32bit_cp0_register(CP0_STATUS) & 0xffff00ff;
		write_32bit_cp0_register(CP0_STATUS, status | irq_mask[0]);
		*VR41XX_MSYSINT1REG = irq_mask[1];
		*VR41XX_MSYSINT2REG = irq_mask[2];
#ifdef CONFIG_CPU_VR4181
		*VR41XX_GPINTMSK = irq_mask[3];
#else
		*VR41XX_MGIUINTLREG = irq_mask[3];
		*VR41XX_MGIUINTHREG = irq_mask[4];
#endif
		break;
	case PM_SUSPEND:
		irq_mask[0] = read_32bit_cp0_register(CP0_STATUS) & 0xff00;
		irq_mask[1] = *VR41XX_MSYSINT1REG;
		irq_mask[2] = *VR41XX_MSYSINT2REG;
#ifdef CONFIG_CPU_VR4181
		irq_mask[3] = *VR41XX_GPINTMSK;
#else
		irq_mask[3] = *VR41XX_MGIUINTLREG;
		irq_mask[4] = *VR41XX_MGIUINTHREG;
#endif
		break;
	}
}
#endif // CONFIG_PM

static struct irqaction cascade = { NULL, SA_INTERRUPT, 0, "cascade", NULL, NULL };
static struct irqaction reserved = { NULL, SA_INTERRUPT, 0, "reserved", NULL, NULL };

void __init init_IRQ(void)
{
	set_except_vector(0, vr41xx_handle_irq);

 	// Default all ICU IRQs to off ...
 	*VR41XX_MSYSINT1REG = 0;
 	*VR41XX_MSYSINT2REG = 0;
 
 	// We initialize the level 2 ICU registers to all bits disabled.
 	// After this, these registers are the resposibility of whatever
 	// driver requests the IRQ of the corresponding level 1 ICU bit.
	// (except GIU, where each level 2 bit has its own IRQ)
#ifdef CONFIG_CPU_VR4122
 	*VR41XX_MPCIINTREG = 0;
 	*VR41XX_MSCUINTREG = 0;
 	*VR41XX_MCSIINTREG = 0;
#else
 	*VR41XX_MPIUINTREG = 0;
 	*VR41XX_MAIUINTREG = 0;
 	*VR41XX_MKIUINTREG = 0;
#endif
#ifdef CONFIG_CPU_VR4181
 	*VR41XX_GPINTMSK = 0xffff;
#else
 	*VR41XX_MGIUINTLREG = 0;
 	*VR41XX_MDSIUINTREG = 0;
 	*VR41XX_MGIUINTHREG = 0;
 	*VR41XX_MFIRINTREG = 0;
#endif
	barrier();

//
// NOTE: This may break autodetection on some devices.  If so, an IRQ mask needs
// to be defined in vr41xx-platdep.h to disable some GPIO lines from interrupting,
// or the broken autodetect IRQ needs to be defined explicitly.  This define is
// for documentation purposes, only turn it off for test/debug:
//
#define LET_ALL_GPIO_INPUTS_CAUSE_INTERRUPTS
#ifdef LET_ALL_GPIO_INPUTS_CAUSE_INTERRUPTS
	// Initialize secondary IRQ mask to enable any GPIO line configured as an input.
	// Note that the int won't be active until enabled in the primary mask, too.
#ifdef CONFIG_CPU_VR4181
	{
		unsigned int bits;

		bits = (unsigned int)(*VR41XX_GPMD1REG | 0xaaaa) << 16;
		bits |= *VR41XX_GPMD0REG | 0xaaaa;
		bits &= bits >> 1;
		bits |= 0xcccccccc;
		bits &= bits >> 2;
		bits |= 0xf0f0f0f0;
		bits &= bits >> 4;
		bits |= 0xff00;
		bits &= bits >> 8;
		*VR41XX_SECIRQMASKL = ~bits & DEVICE_IRQ_MASKL;
	}
#else
	*VR41XX_SECIRQMASKL = ~*VR41XX_GIUIOSELL & DEVICE_IRQ_MASKL;
	*VR41XX_SECIRQMASKH = ~*VR41XX_GIUIOSELH & DEVICE_IRQ_MASKH;
#endif
#endif // LET_ALL_GPIO_INPUTS_CAUSE_INTERRUPTS

 	// These don't really add handlers, these IRQs are never reported by the int
 	// vector handler.  What these do is register the IRQ as non-sharable
 	add_irq_action(VR41XX_IRQ_INT0, &cascade);
 	add_irq_action(VR41XX_IRQ_GIU, &cascade);
 	add_irq_action(VR41XX_IRQ_RTCL1, &reserved);
 	add_irq_action(VR41XX_IRQ_RTCL2, &reserved);

#ifdef CONFIG_REMOTE_DEBUG
 	printk("Setting debug traps - please connect the remote debugger.\n");

	set_debug_traps();

	// you may move this line to whereever you want
	breakpoint();
#endif
}

